/*
  EdiNoS - Editeur de NoS
  Copyright (C) 2013-2021 - Luc Poittevin — Tous droits réservés.
  Contact : edinos@free.fr  -  https://www.edinos.org
  
  Ce programme est un logiciel libre ; vous pouvez le redistribuer ou le
  modifier suivant les termes de la “GNU General Public License” telle que
  publiée par la Free Software Foundation : soit la version 3 de cette
  licence, soit (à votre gré) toute version ultérieure.
  
  Ce programme est distribué dans l’espoir qu’il vous sera utile, mais SANS
  AUCUNE GARANTIE : sans même la garantie implicite de COMMERCIALISABILITÉ
  ni d’ADÉQUATION À UN OBJECTIF PARTICULIER. Consultez la Licence Générale
  Publique GNU pour plus de détails.
  
  Vous devriez avoir reçu une copie de la Licence Générale Publique GNU avec
  ce programme ; si ce n’est pas le cas, consultez :
  <http://www.gnu.org/licenses/>. 
 */

package nos.run;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Serializable;
import java.io.StringWriter;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Collections;

import javax.swing.JFileChooser;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlTransient;

import org.apache.commons.io.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import nos.obj.*;
import nos.utils.ToolNos;

@XmlRootElement
public class Moteur implements Serializable {

	public static enum Contexte {
		Lancement, // mode standard
		ControleNoS, // controle syntaxique du nos courant
		Bilan, // calcul de la couverture
		Test, // pour tests unitaires
		Service // pour appel comme service
	}

	public static enum ModeExecutionProc {
		Ignore, // les procédures ne sont ni exécutées, ni simulées
		// (la base de faits est pré-remplie)
		Simule, // les procédures sont toujours simulées
		Execute, // les procédures sont exécutées
		Mixte
		// les procédures sont exécutées lorsqu'elles sont trouvées,
		// simulées sinon
	}

	private static final long serialVersionUID = 1L;

	public static boolean afficheMT;

	@XmlTransient
	public ModeExecutionProc mode;

	@XmlTransient
	public Contexte contexte;

	protected NoS nos;

	protected Memoire baseFaits;

	protected ArchiveMemoire archiveMemoire;

	protected String log;

	protected List<List<String>> actionsAFaire;;

	protected Map<NoS, Integer> assocChoix;

	protected Moteur ancetre; // contient le moteur du nos appelant lors du déclenchement
							// d'un sous-graphe

	private Moteur cible; // pour ReviNoS : contient le moteur du nos de départ
	// (celui qui est à l'origine de la révision).

	protected String nomActionCourante;

	protected final Logger logTechnique;

	private boolean stop = false;

	public Moteur() { // Pour JAXB
		assocChoix = new HashMap<NoS, Integer>();
		ancetre = null;
		actionsAFaire = new ArrayList<List<String>>();
		logTechnique = LoggerFactory.getLogger(this.getClass());
	}

	public Moteur(NoS n, Contexte c, ModeExecutionProc m) {
		init(n);
		contexte = c;
		mode = m;
		logTechnique = LoggerFactory.getLogger(this.getClass());
	}

	protected void afficher(String s) {
		Message m = null;
		m = Edi.nos().getMessage(s);
		if (m == null) {
			System.out.println("Message <" + s + "> non trouvé"); // utile
			// pour Tests
			Edi.nos().montre("Message <" + s + "> non trouvé");
		} else {
			if (contexte == Contexte.Test) {
				// on ne fait rien
				retourne("");
				return;
			}
			if (nos.getBase().contientMessage(s)) {
				nomActionCourante = m.getNom();
				Edi.nos().affiche(nos.getBase().getMessage(s), this);
			} else {
				Edi.nos().montre("Message <" + s + "> non trouvé");
			}
		}
	}

	protected void afficherTexte(String s) {
		nomActionCourante = null; 
		if (contexte == Contexte.Test) {
			// on ne fait rien
			retourne("");
			return;
		} else {
			Edi.nos().affiche(valoriseFaits(s), this);
		}
	}

	public String valoriseFaits(String s) {
		// <s> contient éventuellement des faits à valoriser entre []...
		// Cette méthode valorise ces faits à l'intérieur de <s>

		char crochetOuvrant = '[';
		char crochetFermant = ']';
		boolean entreCrochet = false;
		String retour = "";
		String fait = "";

		for (char c : s.toCharArray()) {
			if (c == crochetOuvrant) {
				entreCrochet = true;
				fait = "";
			} else if (entreCrochet && (c != crochetFermant)) {
				fait = fait + c;
			} else if (entreCrochet && (c == crochetFermant)) {

				String[] tabFaits = fait.split("\\.");
				if (!baseFaits.contientObjet(tabFaits[0])) {
					// on ne fait rien
				} else {
					String r = baseFaits.evalueFait(tabFaits);
					retour = retour.replaceAll("_", " ");
					if (Arrays.asList(tabFaits).contains(r)) {
						// si la valeur de retour correspond à un des nos
						// c'est que la valeur à retourner est vide
						// (pour une propriété typée uniquement)
						if (Edi.nos().contientProprieteTypee(fait)) {
							r = "";
						}
					}
					retour = retour + r;
					retour = retour.replaceAll("_", " ");
					entreCrochet = false;
				}
			} else {
				retour = retour + c;
			}
		}
		return retour;
	}

	public void ajouteFait(String chaine) {
		/*
		 * Au cours de l'exécution d'une action (exécution qui peut avoir été simulée,
		 * le fait <chaine> est renvoyé, qui est ajouté à <baseFaits>
		 */
		logTechnique.info(nomActionCourante + " ajouteFait " + chaine);

		if (chaine != null) {
			baseFaits.ajoute(nomActionCourante, chaine);
		}
	}

	protected void ajouter(String chaine) {
		if (chaine != null) {
			baseFaits.ajoute("[]", chaine);

		}
		continueActions();
	}

	protected void aller(String nomNos, String p) {

		if (p.length() == 0) {
			p = "0";
		}
		if (nos.getBase().contientNoS(nomNos)) {
			NoS n = nos.getBase().getNoS(nomNos);
			Integer i = Integer.valueOf(p); 
			if (assocChoix.containsKey(n)) {
				if (assocChoix.get(n).compareTo(i) > 0) {
					assocChoix.put(n, i);
				}
			} else {
				assocChoix.put(n, i);
			}
		} else {
			Edi.nos().message("Le nos <\" + nomNos + \"> n'a pas été trouvé...");
		}
	}

	public void back() {

		String action = getDerniereAction();
		// La nouvelle action a été identifiée : il s'agit d'<action>

		if (action != null) {
			// positionne nos courant et maj de la base de Faits :
			MemoireArchivee archive = archiveMemoire.getMemoireArchivee(action);
			nos = Edi.nos().getBaseCourante().getNoS(archive.getNoS());

			// supprime la dernière archive

			archiveMemoire.retireDerniereArchive();

			// maj la base de Faits :
			if (getDerniereArchive() != null) {
				baseFaits = getDerniereArchive().getMemoire().duplique();

			} else {
				baseFaits = new Memoire();
			}

			baseFaits.ajouteParcours(nos.getNom());

			actionsAFaire = ToolNos.asStructureSyntaxe(nos.getActions());

			continueActions();
		}
	}


	public String chaineResultatPlusAction() {
		// "*** renvoie une chaine présentant l'historique ***"
		// "*** enrichie de l'action du nos courant ***"

		String s = baseFaits.ligneHistorique();
		// On ajoute une virgule en fin de chaine :
		if (s.length() > 0) {
			s = s + ",";
		}
			s = s + " "+nos.getNomPremiereAction() + ".";

		return s;
	}

	public String chaineResultatVirgule() {
		// "*** renvoie une chaine présentant l'historique ***"

		String s = baseFaits.ligneHistorique();
		// On ajoute une virgule en fin de chaine :
		if (s.length() > 0) {
			s = s + ",";
		}
		return s;
	}

	protected NoS choixSuivant() {
		// "*** l'association <_assocChoix> a été remplie:
		// "*** il faut renvoyer un nos suivant ou null
		// "*** en fonction de <_assocChoix>

		if (assocChoix.isEmpty()) {
			return null;
		}

		List<Integer> priorites = new ArrayList<Integer>();
		priorites.addAll(assocChoix.values());
		Collections.sort(priorites);
		Integer pr = (Integer) priorites.get(priorites.size() - 1);
		if (pr.intValue() == 0) {
			// "*** uniquement une collection contenant des nodules ***"
			// "*** de priorite 0 : on choisit le premier ... ***"


			return assocChoix.keySet().iterator().next();

		} else {
			// "*** la collection <_assocChoix> ne contient pas que des 0 ***"
			// "*** on choisit le nodule prioritaire, càd de priorite ***"
			// "*** la plus faible ***"
			int minPr = 100;
			NoS nosChoisi = null;
			for (NoS nos : assocChoix.keySet()) {
				Integer p = assocChoix.get(nos);
				if ((p.intValue() > 0) & (p.intValue() < minPr)) {
					minPr = p.intValue();
					nosChoisi = nos;
				}
			}

			return nosChoisi;
		}
	}

	protected void continueActions() {
		if (actionsAFaire.isEmpty()) {
			this.suiteDeclenche();
		} else {
			this.executeActionSuivante();
		}
	}

	protected void declenche(NoS n) {
		// utiliser NoS>>declencheAvec si Contexte.Lancement

		nos = n;
		assocChoix = new HashMap<NoS, Integer>();

		executeActions();
	}

	public void declencheAvec(Memoire mem) {

		archiveMemoire = new ArchiveMemoire();
		archiveMemoire.setMemoiresArchivees(new ArrayList<MemoireArchivee>());
		declencheAvec(mem, archiveMemoire);
	}

	public void declencheAvec(Memoire mem, ArchiveMemoire archive) {
		baseFaits = mem;
		baseFaits.ajouteParcours(nos.getNom());
		archiveMemoire = archive;


		this.executeActions();
	}

	protected void declencher(String s) {
		NoS n = null;
		n = Edi.nos().getNoS(s);
		if (n == null) {
			System.out.println("NoS <" + s + "> non trouvé");
			Edi.nos().montre("NoS <" + s + "> non trouvé");
		} else {
			n.declencheAvec(baseFaits, this, cible, contexte, mode, log, archiveMemoire);
		}
	}

	protected void demander(String s) {
		Question q = null;
		q = Edi.nos().getQuestion(s);
		if (q == null) {
			System.out.println("Question <" + s + "> non trouvée"); // utile
			// pour Tests
			Edi.nos().montre("Question <" + s + "> non trouvée");
		} else {
			if (contexte == Contexte.Test) {
				// on ne fait rien
				retourne("");
				return;
			}
			if (nos.getBase().contientQuestion(s)) {
				if (baseFaits.actionDejaExecutee(s)) {
					Edi.nos().montre("Question <" + s + "> déjà posée");
					continueActions();
				} else {
					nomActionCourante = s;
					Edi.nos().poseQuestion(nos.getBase().getQuestion(s), this);
				}
			} else {
				Edi.nos().montre("Question <" + s + "> non trouvée");
			}
		}
	}


	private void enregistreFichier(String s) {
		String extFic = "xml";
		String nomFic = "export" + getNoS().getNom();

		JFileChooser filechoose = new JFileChooser();
		// Créer un JFileChooser
		// Le répertoire source du JFileChooser est le répertoire d’où est
		// lancé notre programme
		filechoose.setSelectedFile(new File(nomFic + "." + extFic));
		// Le bouton pour valider l’enregistrement portera la mention
		// Enregistrer
		int resultatEnregistrer = filechoose.showSaveDialog(filechoose);
		if (resultatEnregistrer == JFileChooser.APPROVE_OPTION)
		// Si l’utilisateur clique sur le bouton Enregistrer
		{
			String monFichier = new String(filechoose.getSelectedFile().toString());
			// Récupérer le nom du fichier qu’il a spécifié
			if (monFichier.endsWith("." + extFic)) {
				;
			}
			// Si ce nom de fichier finit par extFic, ne rien faire et
			// passer à la suite
			else {
				monFichier = monFichier + "." + extFic;
			}
			{
				try {
					OutputStreamWriter out = new OutputStreamWriter(new FileOutputStream(monFichier), "iso-8859-1");
					// Mettre le flux en tampon (en cache)
					out.write(s); //
					// Balancer dans le flux le contenu de la zone de texte
					out.close(); // Fermer le flux (c’est
					// toujours mieux de le fermer explicitement)
				} catch (IOException er) {
					er.printStackTrace();
				}
			}
		}
	}

	protected void evaluer(String chaine) {
		Procedure p = null;
		p = Edi.nos().getProc(chaine);
		if (p == null) {
			// utile pour tests :
			System.out.println("Procédure <" + chaine + "> non trouvée");
			Edi.nos().montre("Procédure <" + chaine + "> non trouvée");
		} else {
			if (baseFaits.actionDejaExecutee(chaine)) {
				Edi.nos().montre("Procédure <" + chaine + "> déjà exécutée");
				continueActions();
				return;
			}
			if (mode == ModeExecutionProc.Ignore) {
				// on ne fait rien
				retourne("");
			} else if (mode == ModeExecutionProc.Simule) {
				simuleProc(p);
			} else if ((mode == ModeExecutionProc.Execute) || (mode == ModeExecutionProc.Mixte)) {
				Method method = null;
				boolean methodeTrouvee = false;

				List<String> mesClasses = Arrays.asList("Interface", "InterfaceRevi", "InterfaceEdiNoS");

				for (String s : mesClasses) {
					try {
						method = Class.forName("nos.run." + s).getMethod(chaine, Moteur.class);
					} catch (SecurityException e1) {
						e1.printStackTrace();
						continue;
					} catch (NoSuchMethodException e1) {
						continue;
					} catch (ClassNotFoundException e1) {
						continue;
					}
					methodeTrouvee = true;
				}


				if (!methodeTrouvee) {
					if (mode == ModeExecutionProc.Mixte) {
						Edi.nos().montreAvertissement("La procédure <" + chaine
								+ "> n'a pas été trouvée dans la classe Interface : elle a été simulée");
						simuleProc(p);
						return;
					} else { // ModeExecutionProc.Execute
						Edi.nos().montreAvertissement(
								"La procédure <" + chaine + "> n'a pas été trouvée dans la classe Interface.");
						retourne("");
						return;
					}
				}

				try {
					nomActionCourante = chaine;
					baseFaits.ajouteProc(nomActionCourante);
					Object parametres[] = { this };
					method.invoke(null, parametres);
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
					return;
				} catch (IllegalAccessException e) {
					e.printStackTrace();
					return;
				} catch (InvocationTargetException e) {
					String s = "Erreur dans l'éxécution de l'action <" + chaine + ">...";
					e.printStackTrace();
					interruptionExecutionAction(s);
					return;
				}
			}
		}
	}

	private boolean methodeExisteDansInterface(String chaine) {
		Method method = null;
		boolean methodeTrouvee = false;

		List<String> mesClasses = Arrays.asList("Interface", "InterfaceRevi", "InterfaceEdiNoS");

		for (String s : mesClasses) {
			try {
				method = Class.forName("nos.run." + s).getMethod(chaine, Moteur.class);
			} catch (SecurityException e1) {
				e1.printStackTrace();
				continue;
			} catch (NoSuchMethodException e1) {
				continue;
			} catch (ClassNotFoundException e1) {
				continue;
			}
			methodeTrouvee = true;
		}
		if (!methodeTrouvee) {
			for (String s : mesClasses) {
				try {
					method = Class.forName("nos.run." + s).getMethod(chaine, Moteur.class, String.class);
				} catch (SecurityException e1) {
					e1.printStackTrace();
					continue;
				} catch (NoSuchMethodException e1) {
					continue;
				} catch (ClassNotFoundException e1) {
					continue;
				}
				methodeTrouvee = true;
			}
		}
		return methodeTrouvee;
	}

	public void executeActions() {
		// Execution les actions :
		
		String actions = ToolNos.retireCommentaires(nos.getActions());
		if (actions.length() > 0) {
			actionsAFaire = ToolNos.asStructureSyntaxe(actions);
		} 
		
		this.continueActions();
	}

	protected void executeActionSuivante() {
		// declenche l'execution de l'action
		// suivante dans <actionsAFaire>

		List<String> tab = actionsAFaire.remove(0);
		String instruction = tab.get(0);

		if (instruction.equals("evaluer:")) {
			if (tab.size() == 2) {
				this.evaluer(tab.get(1));
			} else if (tab.size() == 3) {
				this.evaluer(tab.get(1), tab.get(2));
			} else {
				Edi.nos().message("Nos <" + nos.getNom() + "> : erreur de syntaxe dans action <evaluer>...");
			}
			return;
		}

		if (instruction.equals("afficher:")) {

			if (tab.size() == 1) { 
				this.objetIntrouvable("message");
				return;
			}

			if (tab.get(1).startsWith("\"")) {
				String s = tab.get(1);
				for (int i = 2; i < tab.size(); i++) {
					s = s + " " + tab.get(i);
				}


				this.afficherTexte(s);
				return;
			} else if (tab.size() == 2) {
				this.afficher(tab.get(1));
			} else {
				this.objetIntrouvable("message");
			}
			return;
		}

		if (instruction.equals("demander:")) {
			if (tab.size() == 2) {
				this.demander(tab.get(1));
			} else {
				this.objetIntrouvable("question");
			}
			return;
		}

		if (instruction.equals("declencher:")) {
			if (tab.size() == 2) {
				this.declencher(tab.get(1));
			} else {
				this.objetIntrouvable("nos");
			}
			return;
		}

		if (instruction.equals("ajouter:")) {
			if (tab.size() == 2) {
				this.ajouter(tab.get(1));
			} else if (tab.size() == 3) {
				this.ajouter(tab.get(1) + " " + tab.get(2));
			} else {
				this.objetIntrouvable("procedure");
			}
			return;
		}

		if (instruction.equals("loguer:")) {
			String s = "";
			for (int i = 1; i < tab.size(); i++) {
				s = s + tab.get(i) + " ";
			}
			// On retire le ' au début et à la fin :
			s = s.substring(1, s.length() - 2);
			this.loguer(s);
			return;
		}

		if (this instanceof MoteurVerifie) {
			((MoteurVerifie) this).ajouteErreur("l'action <" + tab.get(0) + "> est inconnue...");
		} else {
			Edi.nos().message("Nos <" + nos.getNom() + "> : l'action <" + tab.get(0) + "> est inconnue...");
		}
	}

	protected void objetIntrouvable(String s) {
		Edi.nos().message("Nos <" + nos.getNom() + "> : " + s + " introuvable, vérifiez la syntaxe...");
	}

	protected void evaluer(String chaine, String par) {
		String param;

		// analyse et traitement du paramètre :
		if (par.charAt(0) == '"') {
			param = par.replaceAll("\"", "");
		} else {
			param = valoriseFaits("[" + par + "]");
		}

		// le reste de la méthode est identique à évaluer(String chaine) à
		// l'exception des
		// appels getmethod() et invoke()

		Procedure p = null;
		p = Edi.nos().getProc(chaine);
		if (p == null) {
			// utile pour tests :
			System.out.println("Procédure <" + chaine + "> non trouvée");
			Edi.nos().montre("Procédure <" + chaine + "> non trouvée");
		} else {
			if (baseFaits.actionDejaExecutee(chaine)) {
				Edi.nos().montre("Procédure <" + chaine + "> déjà exécutée");
				continueActions();
				return;
			}
			if (mode == ModeExecutionProc.Ignore) {
				// on ne fait rien
				retourne("");
			} else if (mode == ModeExecutionProc.Simule) {
				simuleProc(p);
			} else if ((mode == ModeExecutionProc.Execute) || (mode == ModeExecutionProc.Mixte)) {
				Method method = null;
				boolean methodeTrouvee = false;

				List<String> mesClasses = Arrays.asList("Interface", "InterfaceRevi", "InterfaceEdiNoS");

				for (String s : mesClasses) {
					try {
						method = Class.forName("nos.run." + s).getMethod(chaine, Moteur.class, String.class);
					} catch (SecurityException e1) {
						e1.printStackTrace();
						continue;
					} catch (NoSuchMethodException e1) {
						continue;
					} catch (ClassNotFoundException e1) {
						continue;
					}
					methodeTrouvee = true;
				}

				if (!methodeTrouvee) {
					if (mode == ModeExecutionProc.Mixte) {
						Edi.nos().montreAvertissement("La procédure <" + chaine
								+ "> n'a pas été trouvée dans une classe Interface : elle a été simulée");
						simuleProc(p);
						return;
					} else { // ModeExecutionProc.Execute
						Edi.nos().montreAvertissement("Erreur : la procédure <" + chaine
								+ "> n'a pas été trouvée dans une classe Interface.");
						retourne("");
						return;
					}
				}

				try {
					nomActionCourante = chaine;
					baseFaits.ajouteProc(nomActionCourante);
					Object parametres[] = { this, param };
					method.invoke(null, parametres);
					// sans paramètre :
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
					return;
				} catch (IllegalAccessException e) {
					e.printStackTrace();
					return;
				} catch (InvocationTargetException e) {
					String s = "Erreur dans l'éxécution de l'action <" + chaine + ">...";
					e.printStackTrace();
					interruptionExecutionAction(s);
					return;
				}
			}
		}
	}

	protected void executeChoix(String regles) {
		List<List<String>> liste = new ArrayList<List<String>>();
		List<String> l;


		liste = ToolNos.asStructureSyntaxe(ToolNos.retireCommentaires(regles));

		if (!liste.isEmpty()) {
			for (int i = 0; i < liste.size(); i++) {
				l = (List<String>) liste.get(i);
				if (l.get(0).equals("si:")) {
					String formule, nomNos = "", p = "";
					int ind = l.indexOf("alorsAller:");
					if (ind > 0) {
						nomNos = (String) l.get(ind + 1);
					}
					ind = l.indexOf("priorite:");
					if (ind > 0) {
						p = (String) l.get(ind + 1);
					}
					formule = ToolNos.analyseFormule(l);
					siAlorsAller(formule, nomNos, p);

				} else if (l.get(0).equals("aller:")) {
					String p = "", nomNos = "";
					if (l.contains("priorite:")) {
						p = (String) l.get(3);
					}
					if (l.size() == 1) { 
						if (this instanceof MoteurVerifie) {
							((MoteurVerifie) this).ajouteErreur("la règle <" + regles + "> est incorrecte...");
						} else {
							Edi.nos().message(
									"Nos <" + nos.getNom() + "> : la règle <" + regles + "> est incorrecte...");
						}
						return;
					}
					nomNos = (String) l.get(1);
					aller(nomNos, p);

				} else if (l.get(0).equals("retourner:")) {
					String nomNos = (String) l.get(1);
					retourner(nomNos);

				} else if (l.get(0).isEmpty()) { 
					// cas d'une règle avec juste un commentaire...
					// on ne fait rien
				} else {
					if (this instanceof MoteurVerifie) {
						((MoteurVerifie) this).ajouteErreur("la règle <" + regles 
								+ "> est incorrecte...");
					} else {
						Edi.nos().message("Nos <" + nos.getNom() + "> : la règle <" + regles
								+ "> est incorrecte...");
					}
				}
			}
		}
	}

	protected void finit() {
		

		if (contexte == Contexte.Lancement) {
			if ((cible != null) && (cible.getClass().getName().equals("nos.run.Moteur"))) {
				// on ne montre pas la situation pour les nos de Revinos...
				Edi.nos().montreInfo("fin Revinos sur " + nos.getNom());
			} else {

				Edi.nos().afficheFinLancement(this);
			}
		}
	}

	private String génèreXML() {
		String result;
		StringWriter sw = new StringWriter();
		try {
			JAXBContext context = JAXBContext.newInstance(this.getClass());
			Marshaller marshaller = context.createMarshaller();
			marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);

			marshaller.setProperty(Marshaller.JAXB_ENCODING, "iso-8859-1");
			marshaller.marshal(this, sw);
			result = sw.toString();
		} catch (JAXBException e) {
			e.printStackTrace();
			Edi.nos().message("Erreur lors de la génération du fichier trace...");
			return null;
		}
		return result;
	}

	@XmlTransient
	public List<List<String>> getActionsAFaire() {
		return actionsAFaire;
	}

	@XmlTransient
	public Moteur getAncetre() {
		return ancetre;
	}

	@XmlElement
	public Memoire getBaseFaits() {
		return baseFaits;
	}

	public Map<NoS, Integer> getChoixPossibles() {
		return assocChoix;
	}

	@XmlTransient
	public Moteur getCible() {
		return cible;
	}

	private MemoireArchivee getDerniereArchive() {

		MemoireArchivee derniereArchive = null;
		List<MemoireArchivee> l = archiveMemoire.getMemoiresArchivees();
		if (l.size() > 0) {
			derniereArchive = l.get(l.size() - 1);
		}
		return derniereArchive;
	}

	private String getDerniereAction() {

		String derniereAction = null;

		List<MemoireArchivee> l = archiveMemoire.getMemoiresArchivees();

		if (l.size() > 0) {
			derniereAction = l.get(l.size() - 1).getAction();
		}
		if ((derniereAction == null) || (derniereAction.contentEquals("afficher"))) {
//			l'action "afficher" est enregistrée
//			dans la BF au moment de l'affichage du message
//			et non après le choix de l'utilisateur :
//			il faut la retirer...
			archiveMemoire.retireDerniereArchive(); 
		}
		l = archiveMemoire.getMemoiresArchivees();

		for (int i = l.size() - 1; i >= 0; i--) {
			MemoireArchivee a = l.get(i);
			if (!Edi.nos().contientProc(a.getAction())) {
				// ce n'est pas une procédure
				return a.getAction();
			} else if ((Edi.nos().contientProc(a.getAction()) && (mode == ModeExecutionProc.Simule))) {
				// c'est une procédure et on est en mode simulation
				return a.getAction();
			} else if ((Edi.nos().contientProc(a.getAction()) && (!methodeExisteDansInterface(a.getAction())))) {
				// c'est une procédure qui n'existe pas dans les classes interface
				// donc elle est simulée
				return a.getAction();
			} else {
				archiveMemoire.retireDerniereArchive();
			}
		}

		return derniereAction;
	}


	public String getLog() {
		return log;
	}

	public String getNomActionCourante() {
		return nomActionCourante;
	}

	@XmlElement
	public NoS getNoS() {
		return nos;
	}

	protected void init(NoS n) {
		nos = n;
		assocChoix = new HashMap<NoS, Integer>();
		ancetre = null;
		actionsAFaire = new ArrayList<List<String>>();
	}

	protected void interruptionExecutionAction(String texte) {
		logTechnique.error(texte);
		if ((cible != null) && (cible.getClass().getName().equals("nos.run.Moteur"))) {
			Edi.nos().message("Une erreur est survenue dans Revinos. La révision" + " a été interrompue...");

		} else {
			Edi.nos().message("Nos <" + nos.getNom() + "> : " + texte);
		}
	}

	protected void loguer(String s) {
		if (contexte == Contexte.Test) {
			// on ne fait rien
			retourne("");
			return;
		} else {
			if (log == null) {
				log = s;
			} else {
				log = log + "\n" + s;
			}
		}
		continueActions();
	}


	public void retourne(String chaine) {
		/*
		 * Après exécution d'une action (exécution qui peut avoir été simulée, le fait
		 * <chaine> est renvoyé, qui est ajouté à <baseFaits>
		 */

		if (nomActionCourante != null) 
		{
			logTechnique.info(nomActionCourante + " retourne " + chaine);
		}

		if (chaine != null) {
			baseFaits.ajoute(nomActionCourante, chaine);
		}

		MemoireArchivee m = new MemoireArchivee();
		m.setAction(nomActionCourante);
		m.setMemoire(baseFaits.duplique());
		m.setNoS(nos.getNom());
		archiveMemoire.ajoute(m);

		continueActions();
	}

	private void retourner(String nomNos) {
		// TODO A implémenter
	}

	public void sauve() {
		String result = génèreXML();
		if (result != null) {
			enregistreFichier(result);
		}
	}

	public void sauve(File f) {
		String result = génèreXML();

		if (f != null) {
			try {
				FileUtils.writeStringToFile(f, result, "ISO-8859-1");
			} catch (IOException e1) {
				e1.printStackTrace();
			}
		}
	}

	public void setActionsAFaire(List<List<String>> afaire) {
		actionsAFaire = afaire;
	}

	public void setAncetre(Moteur a) {
		this.ancetre = a;
	}

	public void setBaseFaits(Memoire faits) {
		baseFaits = faits;
	}

	public void setCible(Moteur m) {
		cible = m;
	}

	public void setLog(String l) {
		log = l;
	}

	public void setNomActionCourante(String nom) {
		nomActionCourante = nom;
	}

	public void setNos(NoS noS) {
		nos = noS;
	}

	protected void siAlorsAller(String f, String nomNos, String p) {
		if (p.length() == 0) {
			p = "0";
		}

		

		boolean b = false;
		try {
			b = baseFaits.evalueFormule(f);
		} catch (NoSException e) {
			e.printStackTrace();
			Edi.nos().message("Anomalie dans l'évaluation de la formule.");
		}
		if (b) {
			if (nos.getBase().contientNoS(nomNos)) {
				NoS n = nos.getBase().getNoS(nomNos);
				Integer i = Integer.valueOf(p); 
				if (assocChoix.containsKey(n)) {
					if (assocChoix.get(n).compareTo(i) > 0) {
						assocChoix.put(n, i);
					}
				} else {
					assocChoix.put(n, i);
				}
			} else {
				Edi.nos().montre("Le nos <\" + nomNos + \"> n'a pas été trouvé...");
			}
		}
	}

	private void simuleProc(Procedure p) {
		nomActionCourante = p.getNom();
		Edi.nos().simuleProc(p, this);
	}

	protected void suiteDeclenche() {
		if ((nos.getRegles().length() > 0) && (!stop)) {
			this.executeChoix(nos.getRegles());
		}

		NoS suivant = choixSuivant();
		
		
		// décision finale :
		if (suivant != null) {
			if ((contexte == Contexte.Lancement)) {
				suivant.declencheAvec(baseFaits, ancetre, cible, contexte, mode, log, archiveMemoire);
			} else {
				this.declenche(suivant);
			}
		} else { // pas de suivant
			if (ancetre == null) {
				this.finit();
			} else {
				this.retourDeclenche();
			}
		}
	}

	protected void retourDeclenche() {
		ancetre.setBaseFaits(baseFaits);
		ancetre.continueActions();
	}

	protected void stoppe() {
		// arret du moteur demandé : on vide les actions
		// à faire et on positionne le booleen <stop> de manière à ne plus
		// exécuter d'action et ne plus déclencher de nos suivant
		actionsAFaire.clear();
		ancetre = null;
		stop = true;
	}

	public void traceAffichageMessage(String nomMessage) {
		// Enregistre dans la mémoire de Travail
		// que le message <nomMessage> a été affiché à l'utilisateur

		String nomAction = "afficher";
		baseFaits.ajoute(nomAction, nomMessage);

		MemoireArchivee m = new MemoireArchivee();
		m.setAction(nomAction);
		m.setMemoire(baseFaits.duplique());
		m.setNoS(nos.getNom());
		archiveMemoire.ajoute(m);
	}
}
